"""Test MIP files"""

import copy
import dateutil

from trailblazer.mip import files


def test_get_case_from_config_when_case_id():
    """Test getting the case id from mip config"""
    # GIVEN a case id
    config = {"case_id": "crazygoat"}

    # WHEN getting case from config dict
    case = files.get_case_from_config(config=config)

    # THEN return case
    assert case == "crazygoat"


def test_get_case_from_config_when_family_id():
    """Test getting the case id from mip config"""
    # GIVEN a case id
    config = {"family_id": "crazygoat"}

    # WHEN getting case from config dict
    case = files.get_case_from_config(config=config)

    # THEN return case
    assert case == "crazygoat"


def test_get_case_from_config_when_no_id():
    """Test getting the case id from mip config"""
    # GIVEN a faulty case id
    config = {"not_a_case_id": "anonymous"}

    # WHEN getting case from config dict
    case = files.get_case_from_config(config=config)

    # THEN return case
    assert case == None


def test_parse_config(files_raw) -> dict:
    """
    Args:
    files_raw (dict): With dicts from files
    """

    # GIVEN config data of a "sharp" run (not dry run)
    config_raw = files_raw["config"]

    # WHEN parsing the MIP output config
    config_data = files.parse_config(config_raw)

    # THEN it should work
    assert isinstance(config_data, dict)

    # THEN it should work and the dry run check should be "yes"
    assert config_data["is_dryrun"] is True

    # ... and should include all samples
    assert len(config_raw["analysis_type"]) == len(config_data["samples"])

    # Build dict for return data
    config_test_data = {
        "config_path": "/path_to/cases/case/analysis/case_config.yaml",
        "email": "test.person@test.se",
        "case": "case",
        "log_path": "tests/fixtures/case/mip_2019-07-04T10:47:15.log",
        "sampleinfo_path": "/path_to/cases/case/analysis/case_qc_sample_info.yaml",
        "out_dir": "/path_to/cases/case/analysis",
        "priority": "normal",
    }

    # Check returns from def
    for key, value in config_test_data.items():
        assert config_data[key] == value


def test_parse_sampleinfo_light(files_raw: dict):
    """
    Args:
    files_raw (dict): With dicts from files
    """

    # GIVEN sampleinfo input from a finished analysis
    sampleinfo_raw = files_raw["sampleinfo"]

    # WHEN parsing the MIP output sampleinfo
    sampleinfo_data = files.parse_sampleinfo_light(sampleinfo_raw)

    # THEN it should work
    assert isinstance(sampleinfo_data, dict)

    # THEN it should mark the run as "finished"
    assert sampleinfo_data["is_finished"] is True

    # THEN date should be set
    assert sampleinfo_data["date"] == dateutil.parser.parse("2019-07-05T10:12:03")

    # THEN version should be set
    assert sampleinfo_data["version"] == "v7.1.0"


def test_get_rank_model_version(files_raw: dict, mip_meta_data: dict):
    """Test getting the rank model from sample_info file"""

    # GIVEN sampleinfo input from a finished analysis
    sampleinfo_raw = files_raw["sampleinfo"]

    # WHEN getting the rank_model
    rank_model_version = files.get_rank_model_version(
        sample_info=sampleinfo_raw, rank_model_type="rank_model", step="genmod"
    )

    # THEN the rank model version should be returned
    assert rank_model_version == mip_meta_data["RANK_MODEL_VERSION"]


def test_get_rank_model_version_with_program(files_raw: dict, mip_meta_data: dict):
    """Test getting the rank model from sample_info file with program key"""

    # GIVEN sampleinfo input from a finished analysis
    sampleinfo_raw = copy.copy(files_raw["sampleinfo"])

    # Using old MIP style with program instead of recipe
    sampleinfo_raw["program"] = sampleinfo_raw.pop("recipe")

    # WHEN getting the rank_model
    rank_model_version = files.get_rank_model_version(
        sample_info=sampleinfo_raw, rank_model_type="rank_model", step="genmod"
    )

    # THEN the rank model version should be returned
    assert rank_model_version == mip_meta_data["RANK_MODEL_VERSION"]


def test_get_case_from_sampleinfo_when_case():
    """Test getting the case id from mip sampleinfo file"""
    # GIVEN a case id
    config = {"case": "crazygoat"}

    # WHEN getting case from config dict
    case = files.get_case_from_sampleinfo(sample_info=config)

    # THEN return case
    assert case == "crazygoat"


def test_get_case_from_sampleinfo_when_family():
    """Test getting the case id from mip sampleinfo file"""
    # GIVEN a case id
    config = {"family": "crazygoat"}

    # WHEN getting case from config dict
    case = files.get_case_from_sampleinfo(sample_info=config)

    # THEN return case
    assert case == "crazygoat"


def test_get_case_from_sampleinfo_when_no_id():
    """Test getting the case id from mip config"""
    # GIVEN a faulty case id
    config = {"not_a_case": "anonymous"}

    # WHEN getting case from config dict
    case = files.get_case_from_sampleinfo(sample_info=config)

    # THEN return case
    assert case == None


def test_parse_sampleinfo(files_raw: dict):
    """
    Args:
    files_raw (dict): With dicts from files
    """

    # GIVEN sampleinfo input from a finished analysis
    sampleinfo_raw = files_raw["sampleinfo"]

    # WHEN parsing the MIP output sampleinfo
    sampleinfo_data = files.parse_sampleinfo(sampleinfo_raw)

    # THEN it should work
    assert isinstance(sampleinfo_data, dict)

    # THEN it should mark the run as "finished"
    assert sampleinfo_data["is_finished"] is True

    # More in-depth testing
    # Family data
    # Build dict for family return data
    sampleinfo_test_data = {
        "case": "case",
        "genome_build": "grch37",
        "version": "v7.1.0",
    }

    # Check returns from def 1
    for key, value in sampleinfo_test_data.items():
        assert sampleinfo_data[key] == value

    # Sample data
    # Build dict for sample return data
    sampleinfo_test_sample_data = {"id": "mother"}

    # Check returns from def 2
    for key, value in sampleinfo_test_sample_data.items():
        for sample_data in sampleinfo_data["samples"]:

            if sample_data["id"] != sampleinfo_test_sample_data["id"]:
                continue

            assert sample_data[key] == value

    assert len(sampleinfo_raw["analysis_type"]) == len(sampleinfo_data["samples"])


def test_get_plink_samples(files_raw: dict):
    """ Test get plink sexcheck from qc_metrics"""

    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # WHEN parsing plink output in qc_metrics
    plink_samples = files.get_plink_samples(metrics=qcmetrics_raw)

    expected_plink_samples = {"child": "male", "father": "male", "mother": "female"}
    # THEN the family memebers and their gender should be returned
    assert plink_samples == expected_plink_samples


def test_get_plink_samples_when_program(files_raw: dict):
    """ Test get plink sexcheck from qc_metrics using program key"""

    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = copy.copy(files_raw["qcmetrics"])

    # Using old MIP style with program instead of recipe
    qcmetrics_raw["program"] = qcmetrics_raw.pop("recipe")

    # WHEN parsing plink output in qc_metrics
    plink_samples = files.get_plink_samples(metrics=qcmetrics_raw)

    expected_plink_samples = {"child": "male", "father": "male", "mother": "female"}
    # THEN the family members and their gender should be returned
    assert plink_samples == expected_plink_samples


def test_set_bamstats_metrics_single_bamstat():
    """Test setting bam stats from qc_metrics when single bamstat entry"""
    # GIVEN 2 bamstats file metrics for a sample
    file_metric = {"bamstats": {"raw_total_sequences": 2, "reads_mapped": 1}}
    sample_data = {}

    # WHEN setting bamstats metrics
    sample_data = files.set_bamstats_metrics(
        file_metrics=file_metric, sample_data=sample_data
    )

    expected_sample_data = {"reads": 2, "total_mapped": 1}
    # THEN return bam stats metrics in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_bamstats_metrics_mutiple_bamstat():
    """Test setting bam stats from qc_metrics when multiple bamstat entry"""
    # GIVEN 2 bamstats file metrics for a sample
    sample_metric = {
        "file_1": {"bamstats": {"raw_total_sequences": 2, "reads_mapped": 1}},
        "file_2": {"bamstats": {"raw_total_sequences": 1, "reads_mapped": 2}},
    }
    sample_data = {}

    for file_metric in sample_metric:
        # WHEN setting bamstats metrics
        sample_data = files.set_bamstats_metrics(
            file_metrics=sample_metric[file_metric], sample_data=sample_data
        )

    expected_sample_data = {"reads": 3, "total_mapped": 3}
    # THEN return bam stats metrics in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_chanjo_sexcheck_metrics(files_raw: dict):
    """Test setting gender from file metric for chanjo_sexcheck"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. markduplicates
    sample_metrics = qcmetrics_raw["sample"]["father"][
        "father_lanes_1_sorted_md_brecal_sex"
    ]
    sample_data = {}

    # WHEN setting sample metric for file
    sample_data = files.set_chanjo_sexcheck_metrics(
        file_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {"predicted_sex": "male"}
    # THEN return predicted sex in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_collecthsmetrics_metrics(files_raw: dict, mip_meta_data: dict):
    """Test setting hsmetrics metrics from file metric for collecthsmetrics"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. markduplicates
    sample_metrics = qcmetrics_raw["sample"]["father"][
        "father_lanes_1_sorted_md_brecal_collecthsmetrics"
    ]
    sample_data = {}

    # WHEN setting sample metric for file
    sample_data = files.set_collecthsmetrics_metrics(
        file_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {
        "at_dropout": mip_meta_data["FATHER_AT_DROPOUT"],
        "completeness_target": {
            10: mip_meta_data["FATHER_PCT_TARGET_BASES_10X"],
            20: mip_meta_data["FATHER_PCT_TARGET_BASES_20X"],
            50: mip_meta_data["FATHER_PCT_TARGET_BASES_50X"],
            100: mip_meta_data["FATHER_PCT_TARGET_BASES_100X"],
        },
        "gc_dropout": mip_meta_data["FATHER_GC_DROPOUT"],
        "target_coverage": mip_meta_data["FATHER_MEAN_TARGET_COVERAGE"],
    }
    # THEN return hs metrics in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_collectmultiplemetrics_metrics(files_raw: dict, mip_meta_data: dict):
    """Test setting multiple metrics from file metric for collectmultiplemetrics"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. collectmultiplemetrics
    sample_metrics = qcmetrics_raw["sample"]["father"][
        "father_lanes_1_sorted_md_brecal_collectmultiplemetrics"
    ]
    sample_data = {}

    # WHEN setting sample metric for file
    sample_data = files.set_collectmultiplemetrics_metrics(
        file_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {"strand_balance": mip_meta_data["FATHER_STRAND_BALANCE"]}
    # THEN return strand balance in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_collectmultiplemetricsinsertsize_metrics(
    files_raw: dict, mip_meta_data: dict
):
    """Test setting multiple metrics from file metric for collectmultiplemetricsinsertsize"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. collectmultiplemetricsinsertsize
    sample_metrics = qcmetrics_raw["sample"]["father"][
        "father_lanes_1_sorted_md_brecal_collectmultiplemetrics"
    ]
    sample_data = {}

    # WHEN setting sample metric for file
    sample_data = files.set_collectmultiplemetricsinsertsize_metrics(
        file_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {
        "median_insert_size": mip_meta_data["FATHER_MEDIAN_INSERT_SIZE"],
        "insert_size_standard_deviation": mip_meta_data["FATHER_STANDARD_DEVIATION"],
    }
    # THEN return insert size metrics in sample data for file metric
    assert sample_data == expected_sample_data


def test_set_markduplicates_metrics(files_raw: dict, mip_meta_data: dict):
    """Test setting duplicates metrics from file metric for markduplicates"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. markduplicates
    sample_metrics = qcmetrics_raw["sample"]["father"]["father_lanes_1_sorted_md"]
    sample_data = {}

    # WHEN setting sample metric for file
    sample_data = files.set_markduplicates_metrics(
        file_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {"duplicates": mip_meta_data["FATHER_FRACTION_DUPLICATES"]}
    # THEN return duplicates in sample data for file metric
    assert sample_data == expected_sample_data


def test_get_sample_metrics(files_raw: dict, mip_meta_data: dict):
    """Test getting sample data for duplicates from file metrics for markduplicates"""
    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # Isolating to a single file_metric i.e. markduplicates
    sample_metrics = {
        "father_lanes_1_sorted_md": qcmetrics_raw["sample"]["father"][
            "father_lanes_1_sorted_md"
        ]
    }
    sample_data = {}

    # WHEN getting sample metric for file
    sample_data = files.get_sample_metrics(
        sample_metrics=sample_metrics, sample_data=sample_data
    )

    expected_sample_data = {"duplicates": mip_meta_data["FATHER_FRACTION_DUPLICATES"]}
    # THEN return duplicates in sample data for file metric
    assert sample_data == expected_sample_data


def test_parse_qcmetrics(files_raw: dict, mip_meta_data: dict):
    """
    Args:
    files_raw (dict): With dicts from files
    """

    # GIVEN qc metrics input from an analysis
    qcmetrics_raw = files_raw["qcmetrics"]

    # WHEN parsing qc metric data
    qcmetrics_data = files.parse_qcmetrics(qcmetrics_raw)

    # THEN it should return a dict
    assert isinstance(qcmetrics_data, dict)

    # Sample data
    # Build dict for sample return data
    qcmetrics_test_sample_data = {
        "at_dropout": mip_meta_data["MOTHER_AT_DROPOUT"],
        "duplicates": mip_meta_data["MOTHER_FRACTION_DUPLICATES"],
        "id": "mother",
        "insert_size_standard_deviation": mip_meta_data["MOTHER_STANDARD_DEVIATION"],
        "gc_dropout": mip_meta_data["MOTHER_GC_DROPOUT"],
        "mapped": mip_meta_data["MOTHER_MAPPED"],
        "median_insert_size": mip_meta_data["MOTHER_MEDIAN_INSERT_SIZE"],
        "plink_sex": "female",
        "predicted_sex": "female",
        "reads": mip_meta_data["MOTHER_RAW_TOTAL_SEQUENCES"],
        "strand_balance": mip_meta_data["MOTHER_STRAND_BALANCE"],
        "target_coverage": mip_meta_data["MOTHER_MEAN_TARGET_COVERAGE"],
    }

    # Check returns from def
    for key, value in qcmetrics_test_sample_data.items():
        for sample_data in qcmetrics_data["samples"]:

            if sample_data["id"] != qcmetrics_test_sample_data["id"]:
                continue

            assert sample_data[key] == value

    # Sample coverage data
    # Build dict for sample coverage return data
    qcmetrics_test_sample_cov_data = {
        10: mip_meta_data["MOTHER_PCT_TARGET_BASES_10X"],
        20: mip_meta_data["MOTHER_PCT_TARGET_BASES_20X"],
        50: mip_meta_data["MOTHER_PCT_TARGET_BASES_50X"],
        100: mip_meta_data["MOTHER_PCT_TARGET_BASES_100X"],
    }

    # Check returns from def
    for key, value in qcmetrics_test_sample_cov_data.items():
        for sample_data in qcmetrics_data["samples"]:

            if sample_data["id"] != qcmetrics_test_sample_data["id"]:
                continue

            assert sample_data["completeness_target"][key] == value



# -*- coding: utf-8 -*-
from typing import List
import datetime as dt

import alchy
import sqlalchemy as sqa

from trailblazer.mip.config import ConfigHandler
from trailblazer.constants import COMPLETED_STATUS, FAILED_STATUS, ONGOING_STATUSES
from . import models


class BaseHandler:
    User = models.User
    Analysis = models.Analysis
    Job = models.Job
    Info = models.Info

    def setup(self):
        self.create_all()
        # add inital metadata record (for web interface)
        new_info = self.Info()
        self.add_commit(new_info)

    def find_analysis(self, family, started_at, status):
        """Find a single analysis."""
        query = self.Analysis.query.filter_by(
            family=family, started_at=started_at, status=status
        )
        return query.first()

    def find_analyses_with_comment(self, comment):
        """Find a analyses containing comment."""
        analysis_query = self.Analysis.query

        analysis_query = analysis_query.filter(
            self.Analysis.comment.like(f"%{comment}%")
        )
        return analysis_query

    def analyses(
        self,
        *,
        family: str = None,
        query: str = None,
        status: str = None,
        deleted: bool = None,
        temp: bool = False,
        before: dt.datetime = None,
        is_visible: bool = None,
    ):
        """Fetch analyses form the database."""
        analysis_query = self.Analysis.query
        if family:
            analysis_query = analysis_query.filter_by(family=family)
        elif query:
            analysis_query = analysis_query.filter(
                sqa.or_(
                    self.Analysis.family.like(f"%{query}%"),
                    self.Analysis.status.like(f"%{query}%"),
                )
            )
        if status:
            analysis_query = analysis_query.filter_by(status=status)
        if isinstance(deleted, bool):
            analysis_query = analysis_query.filter_by(is_deleted=deleted)
        if temp:
            analysis_query = analysis_query.filter(
                self.Analysis.status.in_(ONGOING_STATUSES)
            )
        if before:
            analysis_query = analysis_query.filter(self.Analysis.started_at < before)
        if is_visible is not None:
            analysis_query = analysis_query.filter_by(is_visible=is_visible)
        return analysis_query.order_by(self.Analysis.started_at.desc())

    def analysis(self, analysis_id: int) -> models.Analysis:
        """Get a single analysis."""
        return self.Analysis.query.get(analysis_id)

    def track_update(self):
        """Update the lastest updated date in the database."""
        metadata = self.info()
        metadata.updated_at = dt.datetime.now()
        self.commit()

    def is_latest_analysis_ongoing(self, case_id: str) -> bool:
        """Check if the latest analysis is ongoing for a case_id"""
        latest_analysis = self.analyses(family=case_id).first()
        if latest_analysis and latest_analysis.status in ONGOING_STATUSES:
            return True
        return False

    def is_latest_analysis_failed(self, case_id: str) -> bool:
        """Check if the latest analysis is failed for a case_id"""
        latest_analysis = self.analyses(family=case_id).first()
        if latest_analysis and latest_analysis.status == FAILED_STATUS:
            return True
        return False

    def get_latest_analysis_status(self, case_id: str) -> str:
        """Get latest analysis status for a case_id"""
        latest_analysis = self.analyses(family=case_id).first()
        return latest_analysis.status

    def is_latest_analysis_completed(self, case_id: str) -> bool:
        """Check if the latest analysis is completed for a case_id"""
        latest_analysis = self.analyses(family=case_id).first()
        if latest_analysis and latest_analysis.status == COMPLETED_STATUS:
            return True
        return False

    def info(self) -> models.Info:
        """Return metadata entry."""
        return self.Info.query.first()

    def add_pending(self, family: str, email: str = None) -> models.Analysis:
        """Add pending entry for an analysis."""
        started_at = dt.datetime.now()
        new_log = self.Analysis(family=family, status="pending", started_at=started_at)
        new_log.user = self.user(email) if email else None
        self.add_commit(new_log)
        return new_log

    def add_user(self, name: str, email: str) -> models.User:
        """Add a new user to the database."""
        new_user = self.User(name=name, email=email)
        self.add_commit(new_user)
        return new_user

    def user(self, email: str) -> models.User:
        """Fetch a user from the database."""
        return self.User.query.filter_by(email=email).first()

    def aggregate_failed(self, since_when: dt.date = None) -> List:
        """Count the number of failed jobs per category (name)."""

        categories = self.session.query(
            self.Job.name.label("name"), sqa.func.count(self.Job.id).label("count")
        ).filter(self.Job.status != "cancelled")

        if since_when:
            categories = categories.filter(self.Job.started_at > since_when)

        categories = categories.group_by(self.Job.name).all()

        data = [
            {"name": category.name, "count": category.count} for category in categories
        ]
        return data

    def jobs(self):
        """Return all jobs in the database."""
        return self.Job.query


class Store(alchy.Manager, BaseHandler, ConfigHandler):
    def __init__(self, uri: str, families_dir: str):
        super(Store, self).__init__(
            config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=models.Model
        )
        self.families_dir = families_dir



"""Tests for the config functionality"""
import pytest

from trailblazer.mip import config as tb_config
from trailblazer.exc import ConfigError


def test_validate_config():
    """Test to validate a correct config"""
    # GIVEN a correct config
    sample = dict(
        sample_id="sample",
        analysis_type="wes",
        father="0",
        mother="0",
        phenotype="affected",
        sex="male",
        expected_coverage=15,
        capture_kit="agilent_sureselect_cre.v1",
    )

    config = dict(
        case="a_case",
        default_gene_panels=["a_panel"],
        samples=[sample],
    )

    # WHEN validating the config
    errors = tb_config.ConfigHandler.validate_config(config)

    # THEN assert the errors is a dict
    assert isinstance(errors, dict)

    # THEN assert there where no errors
    assert errors == {}


def test_validate_config_invalid_analysis_type():
    """Test to validate a invalid config with wrong analysis type"""
    # GIVEN a correct config
    sample = dict(
        sample_id="sample",
        analysis_type="nonexisting",
        father="0",
        mother="0",
        phenotype="affected",
        sex="male",
        expected_coverage=15,
        capture_kit="agilent_sureselect_cre.v1",
    )

    config = dict(
        case="a_case",
        default_gene_panels=["a_panel"],
        samples=[sample],
    )

    # WHEN validating the config
    with pytest.raises(ConfigError):
        # THEN assert a config error is raised
        tb_config.ConfigHandler.validate_config(config)


def test_validate_config_unknown_field():
    """Test to validate a config with a unspecified field.

    This should work since we allow unspecified fields
    """
    # GIVEN a correct config with a extra field
    sample = dict(
        sample_id="sample",
        sample_display_name="ASAMPLENAME",
        analysis_type="wes",
        father="0",
        mother="0",
        phenotype="affected",
        sex="male",
        expected_coverage=15,
        capture_kit="agilent_sureselect_cre.v1",
    )

    config = dict(
        case="a_case",
        default_gene_panels=["a_panel"],
        samples=[sample],
    )

    # WHEN validating the config
    errors = tb_config.ConfigHandler.validate_config(config)

    # THEN assert the errors is a dict
    assert isinstance(errors, dict)

    # THEN assert there where one error that could pass
    assert len(errors) == 1


def test_validate_config_unknown_field_and_missing_sample_id():
    """Test to validate a config with a unspecified field and a missing mandatory field.

    This should raise exception since we do not allow missing mandatory fields
    """
    # GIVEN a config with missing sample_id and an extra field
    sample = dict(
        sample_display_name="ASAMPLENAME",
        analysis_type="wes",
        father="0",
        mother="0",
        phenotype="affected",
        sex="male",
        expected_coverage=15,
        capture_kit="agilent_sureselect_cre.v1",
    )

    config = dict(
        case="a_case",
        default_gene_panels=["a_panel"],
        samples=[sample],
    )

    # WHEN validating the config
    with pytest.raises(ConfigError):
        # THEN assert a config error is raised
        tb_config.ConfigHandler.validate_config(config)


def test_validate_config_unknown_field_and_invalid_analysis_type():
    """Test to validate a config with a unspecified field and a missing mandatory field.

    This should raise exception since we do not allow wrong analysis types
    """
    # GIVEN a config with wrong analysis type and an extra field
    sample = dict(
        sample_id="sample",
        sample_display_name="ASAMPLENAME",
        analysis_type="nonexisting",
        father="0",
        mother="0",
        phenotype="affected",
        sex="male",
        expected_coverage=15,
        capture_kit="agilent_sureselect_cre.v1",
    )

    config = dict(
        case="a_case",
        default_gene_panels=["a_panel"],
        samples=[sample],
    )

    # WHEN validating the config
    with pytest.raises(ConfigError):
        # THEN assert a config error is raised
        tb_config.ConfigHandler.validate_config(config)



# -*- coding: utf-8 -*-
import datetime as dt
from pathlib import Path
import logging
from typing import List

log = logging.getLogger(__name__)


class FastqHandler:
    @staticmethod
    def name_file(
        lane: int,
        flowcell: str,
        sample: str,
        read: int,
        undetermined: bool = False,
        date: dt.datetime = None,
        index: str = None,
    ) -> str:
        """Name a FASTQ file following MIP conventions."""
        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        date_str = date.strftime("%y%m%d") if date else "171015"
        index = index if index else "XXXXXX"
        return f"{lane}_{date_str}_{flowcell}_{sample}_{index}_{read}.fastq.gz"

    def link(self, family: str, sample: str, analysis_type: str, files: List[str]):
        """Link FASTQ files for a sample."""
        root_dir = Path(self.families_dir) / family / analysis_type / sample / "fastq"
        root_dir.mkdir(parents=True, exist_ok=True)
        for fastq_data in files:
            fastq_path = Path(fastq_data["path"])
            fastq_name = self.name_file(
                lane=fastq_data["lane"],
                flowcell=fastq_data["flowcell"],
                sample=sample,
                read=fastq_data["read"],
                undetermined=fastq_data["undetermined"],
            )
            dest_path = root_dir / fastq_name
            if not dest_path.exists():
                log.info(f"linking: {fastq_path} -> {dest_path}")
                dest_path.symlink_to(fastq_path)
            else:
                log.debug(f"destination path already exists: {dest_path}")



# -*- coding: utf-8 -*-


def test_build_command(mip_cli):
    # GIVEN just a config
    mip_config = "tests/fixtures/global_config.yaml"
    case_id = "angrybird"
    # WHEN building the MIP command
    mip_command = mip_cli.build_command(case=case_id, config=mip_config)
    # THEN it should use the correct options
    assert mip_command == [
        mip_cli.script,
        mip_cli.pipeline,
        case_id,
        "--config_file",
        mip_config,
    ]



# -*- coding: utf-8 -*-
import logging
import signal
import subprocess

from trailblazer.exc import MipStartError

LOG = logging.getLogger(__name__)

CLI_OPTIONS = {
    "config": {"option": "--config_file"},
    "priority": {"option": "--slurm_quality_of_service"},
    "email": {"option": "--email"},
    "base": {"option": "--cluster_constant_path"},
    "dryrun": {"option": "--dry_run_all"},
    "gene_list": {"option": "--vcfparser_slt_fl"},
    "max_gaussian": {"option": "--gatk_varrecal_snv_max_gau"},
    "skip_evaluation": {"option": "--qccollect_skip_evaluation"},
    "start_with": {"option": "--start_with_recipe"},
}


class MipCli(object):

    """Wrapper around MIP command line interface."""

    def __init__(self, script, pipeline, conda_env):
        """Initialize MIP command line interface."""
        self.script = script
        self.pipeline = pipeline
        self.conda_env = conda_env

    def __call__(self, config, case, **kwargs):
        """Execute the pipeline."""
        command = self.build_command(case, config, **kwargs)
        LOG.debug(" ".join(command))
        process = self.execute(command)
        if process.returncode != 0:
            raise MipStartError(
                f"Error starting analysis: " f"{subprocess.CalledProcessError.stderr}"
            )
        return process

    def build_command(self, case, config, **kwargs):
        """Builds the command to execute MIP."""
        command = [
            self.script,
            self.pipeline,
            case,
            CLI_OPTIONS["config"]["option"],
            config,
        ]
        for key, value in kwargs.items():
            # enable passing in flags as "False" - shouldn't add command
            if value:
                command.append(CLI_OPTIONS[key]["option"])

                # append value for non-flags
                if value is not True:
                    command.append(value)
        return command

    def execute(self, command):
        """Start a new MIP run."""

        command.insert(0, f"bash -c 'source activate {self.conda_env}; ")
        command.append("'")

        process = subprocess.run(" ".join(command), shell=True, check=True)
        return process



# -*- coding: utf-8 -*-
from copy import deepcopy
import logging
from pathlib import Path

from marshmallow import Schema, fields, validate
import ruamel.yaml

from trailblazer.constants import DEFAULT_CAPTURE_KIT
from trailblazer.exc import ConfigError

LOG = logging.getLogger(__name__)


class SampleSchema(Schema):
    sample_id = fields.Str(required=True)
    analysis_type = fields.Str(
        required=True, validate=validate.OneOf(choices=["wes", "wgs", "tga"])
    )
    father = fields.Str(default="0")
    mother = fields.Str(default="0")
    phenotype = fields.Str(
        required=True,
        validate=validate.OneOf(choices=["affected", "unaffected", "unknown"]),
    )
    sex = fields.Str(
        required=True, validate=validate.OneOf(choices=["male", "female", "unknown"])
    )
    expected_coverage = fields.Float()
    capture_kit = fields.Str(default=DEFAULT_CAPTURE_KIT)


class ConfigSchema(Schema):
    case = fields.Str(required=True)
    default_gene_panels = fields.List(fields.Str(), required=True)
    samples = fields.List(fields.Nested(SampleSchema), required=True)


class SampleSchemaRNA(Schema):
    sample_id = fields.Str(required=True)
    analysis_type = fields.Str(required=True, validate=validate.OneOf(choices=["wts"]))
    father = fields.Str(default="0")
    mother = fields.Str(default="0")
    phenotype = fields.Str(
        required=True,
        validate=validate.OneOf(choices=["affected", "unaffected", "unknown"]),
    )
    sex = fields.Str(
        required=True, validate=validate.OneOf(choices=["male", "female", "unknown"])
    )
    expected_coverage = fields.Float()
    capture_kit = fields.Str(default=DEFAULT_CAPTURE_KIT)


class ConfigSchemaRNA(Schema):
    case = fields.Str(required=True)
    default_gene_panels = fields.List(fields.Str(), required=True)
    samples = fields.List(fields.Nested(SampleSchemaRNA), required=True)


class ConfigHandler:
    def make_config(self, data: dict, pipeline: str = None):
        """Make a MIP config."""
        self.validate_config(data, pipeline)
        config_data = self.prepare_config(data)
        return config_data

    @staticmethod
    def validate_config(data: dict, pipeline: str = None) -> dict:
        """Convert to MIP config format."""
        if pipeline == "mip-rna":
            errors = ConfigSchemaRNA().validate(data)
        else:
            errors = ConfigSchema().validate(data)
        if errors:
            hard_error = False
            for field, messages in errors.items():
                if isinstance(messages, dict):
                    for level, sample_errors in messages.items():
                        try:
                            sample_id = data["samples"][level]["sample_id"]
                        except KeyError:
                            raise ConfigError("missing sample id")

                        for sub_field, sub_messages in sample_errors.items():
                            if sub_messages != ["Unknown field."]:
                                hard_error = True
                            LOG.error(
                                f"{sample_id} -> {sub_field}: {', '.join(sub_messages)}"
                            )
                else:
                    hard_error = True
                    LOG.error(f"{field}: {', '.join(messages)}")
            if hard_error:
                raise ConfigError("invalid config input", errors=errors)
        return errors

    @staticmethod
    def prepare_config(data: dict) -> dict:
        """Prepare the config data."""
        data_copy = deepcopy(data)
        # handle single sample cases with 'unknown' phenotype
        if len(data_copy["samples"]) == 1:
            if data_copy["samples"][0]["phenotype"] == "unknown":
                LOG.info("setting 'unknown' phenotype to 'unaffected'")
                data_copy["samples"][0]["phenotype"] = "unaffected"
        # set the mother/father to '0' if they are not set for a sample
        for sample_data in data_copy["samples"]:
            sample_data["mother"] = sample_data.get("mother") or "0"
            sample_data["father"] = sample_data.get("father") or "0"
            if (
                sample_data["analysis_type"] == "wgs"
                and sample_data.get("capture_kit") is None
            ):
                sample_data["capture_kit"] = DEFAULT_CAPTURE_KIT
        return data_copy

    def save_config(self, data: dict) -> Path:
        """Save a config to the expected location."""
        out_dir = Path(self.families_dir) / data["case"]
        out_dir.mkdir(parents=True, exist_ok=True)
        out_path = out_dir / "pedigree.yaml"
        dump = ruamel.yaml.round_trip_dump(data, indent=4, block_seq_indent=2)
        out_path.write_text(dump)
        return out_path
